library bloc_stream;

import 'dart:async';

import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart';

typedef Action<B extends BlocStream<State>, State> = FutureOr<void> Function(
    B bloc, void Function(State) add);

Stream<State> Function(dynamic) actionToStream<State>(BlocStream<State> bloc) =>
    (action) {
      final controller = StreamController<State>();
      Future.value(action(bloc, controller.add)).catchError((err) {
        print('[${bloc.runtimeType}] Error: $err');
      }).whenComplete(() => controller.close());
      return controller.stream;
    };

abstract class BlocStream<State> extends Stream<State> {
  BlocStream(State initialValue) {
    _state$ = BehaviorSubject.seeded(initialValueOverride ?? initialValue);
    _bindActionsToStates();
  }

  final _actionController = StreamController<dynamic>();
  late BehaviorSubject<State> _state$;

  State? get initialValueOverride => null;
  State get value => _state$.value!;

  @override
  bool get isBroadcast => _state$.isBroadcast;

  @override
  StreamSubscription<State> listen(
    void Function(State)? onData, {
    Function? onError,
    void Function()? onDone,
    bool? cancelOnError,
  }) {
    return _state$.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  /// Add actions to be processed. Implements Sink
  void add<B extends BlocStream<State>>(Action<B, State> action) {
    if (_actionController.isClosed) return;
    _actionController.add(action as dynamic);
  }

  @mustCallSuper
  void willUpdateValue(State current, State next) {}

  /// Transforms a stream of actions to a stream of state.
  /// You can override this to throttle actions for example.
  Stream<State> transformActions(
    Stream<dynamic> actions$,
    Stream<State> Function(dynamic) transform,
  ) =>
      actions$.asyncExpand(transform);

  void _bindActionsToStates() {
    transformActions(
      _actionController.stream,
      actionToStream(this),
    ).listen((newState) {
      if (newState == value) return;
      willUpdateValue(value, newState);
      _state$.add(newState);
    }, onError: (error) {
      _state$.addError(error);
    }, onDone: () {
      _state$.close();
    });
  }

  /// Helper method to reduce boilerplate when composing BLoC's together.
  final _closers = <void Function()>[];
  void cancelOnClose(StreamSubscription subscription,
      [void Function()? onClose]) {
    _closers.add(() {
      subscription.cancel();
      if (onClose != null) onClose();
    });
  }

  @mustCallSuper
  Future<void> close() {
    _closers.forEach((fn) => fn());
    return _actionController.close();
  }
}
